@@ -24,9 +24,16 @@
 
 #include <string.h>
 #include <vte/vte.h>
+#include <gdk/gdkx.h>
 
 #define UNIMPLEMENTED /* g_warning (G_STRLOC": unimplemented") */
 
+typedef struct
+{
+  int tag;
+  int flavor;
+} TagData;
+
 typedef struct
 {
   GSList *url_tags;
@@ -40,7 +47,10 @@ free_vte_data (gpointer data)
 
   vte = data;
 
+  g_slist_foreach (vte->url_tags, (GFunc) g_free, NULL);
   g_slist_free(vte->url_tags);
+
+  g_slist_foreach (vte->skey_tags, (GFunc) g_free, NULL);
   g_slist_free(vte->skey_tags);
   
   g_free (vte);
@@ -115,28 +125,42 @@ terminal_widget_get_padding                (GtkWidget            *widget,
 
 void
 terminal_widget_match_add                  (GtkWidget            *widget,
-					    const char           *regexp)
+					    const char           *regexp,
+                                            int                   flavor)
 {
+  TagData *tag_data;
   VteData *data;
   int tag;
   
   data = g_object_get_data (G_OBJECT (widget), "terminal-widget-data");
 
   tag = vte_terminal_match_add (VTE_TERMINAL (widget), regexp);
-  data->url_tags = g_slist_append (data->url_tags, GINT_TO_POINTER(tag));
+
+  tag_data = g_new0 (TagData, 1);
+  tag_data->tag = tag;
+  tag_data->flavor = flavor;
+
+  data->url_tags = g_slist_append (data->url_tags, tag_data);
 }
 
 void
 terminal_widget_skey_match_add             (GtkWidget            *widget,
-					    const char           *regexp)
+					    const char           *regexp,
+                                            int                   flavor)
 {
+  TagData *tag_data;
   VteData *data;
   int tag;
   
   data = g_object_get_data (G_OBJECT (widget), "terminal-widget-data");
 
   tag = vte_terminal_match_add(VTE_TERMINAL(widget), regexp);
-  data->skey_tags = g_slist_append (data->skey_tags, GINT_TO_POINTER(tag));
+
+  tag_data = g_new0 (TagData, 1);
+  tag_data->tag = tag;
+  tag_data->flavor = flavor;
+
+  data->skey_tags = g_slist_append (data->skey_tags, tag_data);
 }
 
 void
@@ -149,8 +173,9 @@ terminal_widget_skey_match_remove          (GtkWidget            *widget)
 
   for (tags = data->skey_tags; tags != NULL; tags = g_slist_next(tags))
     vte_terminal_match_remove(VTE_TERMINAL(widget),
-   			      GPOINTER_TO_INT(tags->data));
+   			      GPOINTER_TO_INT(((TagData*)tags->data)->tag));
 
+  g_slist_foreach (data->skey_tags, (GFunc) g_free, NULL);
   g_slist_free(data->skey_tags);
   data->skey_tags = NULL;
 }
@@ -158,7 +183,8 @@ terminal_widget_skey_match_remove          (GtkWidget            *widget)
 char*
 terminal_widget_check_match (GtkWidget *widget,
 			     int        column,
-			     int        row)
+			     int        row,
+                             int       *flavor)
 {
   VteData *data;
   GSList *tags;
@@ -169,8 +195,12 @@ terminal_widget_check_match (GtkWidget *widget,
 
   match = vte_terminal_match_check(VTE_TERMINAL(widget), column, row, &tag);
   for (tags = data->url_tags; tags != NULL; tags = g_slist_next(tags))
-    if (GPOINTER_TO_INT(tags->data) == tag)
-      return match;
+    if (GPOINTER_TO_INT(((TagData*)tags->data)->tag) == tag)
+      {
+        if (flavor)
+          *flavor = tag;
+        return match;
+      }
 
   g_free (match);
   return NULL;
@@ -179,7 +209,8 @@ terminal_widget_check_match (GtkWidget *widget,
 char*
 terminal_widget_skey_check_match (GtkWidget *widget,
 				  int        column,
-				  int        row)
+				  int        row,
+                                  int       *flavor)
 {
   VteData *data;
   GSList *tags;
@@ -190,12 +221,15 @@ terminal_widget_skey_check_match (GtkWidget *widget,
 
   match = vte_terminal_match_check(VTE_TERMINAL(widget), column, row, &tag);
   for (tags = data->skey_tags; tags != NULL; tags = g_slist_next(tags))
-    if (GPOINTER_TO_INT(tags->data) == tag)
-      return match;
+    if (GPOINTER_TO_INT(((TagData*)tags->data)->tag) == tag)
+      {
+        if (flavor)
+          *flavor = tag;
+        return match;
+      }
 
   g_free (match);
   return NULL;
-      
 }
 
 void
@@ -520,6 +554,7 @@ terminal_widget_get_scroll_adjustment (GtkWidget *widget)
 
 gboolean
 terminal_widget_fork_command (GtkWidget   *widget,
+                              gboolean     lastlog,
 			      gboolean     update_records,
 			      const char  *path,
 			      char       **argv,
@@ -530,7 +565,7 @@ terminal_widget_fork_command (GtkWidget   *widget,
 {
   *child_pid = vte_terminal_fork_command (VTE_TERMINAL (widget),
 		 			  path, argv, envp, working_dir,
-					  update_records, TRUE, TRUE);
+					  lastlog, update_records, update_records);
 
   if (*child_pid == -1)
     {
@@ -565,10 +600,31 @@ terminal_widget_write_data_to_child (GtkWidget  *widget,
 
 void
 terminal_widget_set_pango_font (GtkWidget                  *widget,
-                                const PangoFontDescription *font_desc)
+				const PangoFontDescription *font_desc,
+				gboolean no_aa_without_render)
 {
   g_return_if_fail (font_desc != NULL);
-  vte_terminal_set_font (VTE_TERMINAL (widget), font_desc);
+
+  if (!no_aa_without_render)
+    vte_terminal_set_font (VTE_TERMINAL (widget), font_desc);
+
+  else
+    {
+      Display *dpy;
+      gboolean has_render;
+      gint event_base, error_base;
+
+      dpy = gdk_x11_display_get_xdisplay (gdk_display_get_default ());
+      has_render = (XRenderQueryExtension (dpy, &event_base, &error_base) &&
+		    (XRenderFindVisualFormat (dpy, DefaultVisual (dpy, DefaultScreen (dpy))) != 0));
+
+      if (has_render)
+	vte_terminal_set_font (VTE_TERMINAL (widget), font_desc);
+      else 
+	vte_terminal_set_font_full (VTE_TERMINAL (widget),
+				    font_desc,
+				    VTE_ANTI_ALIAS_FORCE_DISABLE);
+    }
 }
 
 gboolean
